//---------------------------------------------------------------------------
#include <vcl>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <time.h>
#include <conio.h>

#define MAX(x,y) ((x)<(y) ? (y) : (x))

#pragma hdrstop

struct lista {
	int posicao;
  	struct lista *proximo;  /* ponteiro para a prxima entrada */
  	struct lista *anterior; /* ponteiro para o registro anterior */
};

struct lista_restrita {
	int cidade;
        int demanda;
        double distancia;
  	struct lista_restrita *proximo;  /* ponteiro para a prxima entrada */
  	struct lista_restrita *anterior; /* ponteiro para o registro anterior */
};

int *cria_vetor(int tam);
double *cria_vetor_double(int tam);
void libera_vetor(int *vetor);
double **cria_matriz(int nlinhas, int ncolunas);
void libera_matriz(double **matriz, int nlinhas);
void obter_parametros_prob(char nomearq[], int *n, int *q, int *dx, int *dy);
void le_arq_matriz(char nomearq[], int n, int dx, int dy, double **distancia, int *demanda);

void imprime_solucao(int *s);
float calcula_fo(int *s, double **distancia, int inviabilidade, int penalidade);
double randomico(double min, double max);
void atualiza_melhor_solucao(int *s, int *s_star, int num_objetos);
int calcula_inviabilidade(int *s, int *d, int q);
int calcula_penalidade(int *d, int n);
void inicializa_vetor(int *vetor, int tam);

void insere_LC(struct lista_restrita *i,    /* novo elemento */
               struct lista_restrita **inicio_LC, /* primeiro elemento da lista */
               struct lista_restrita **final_LC);  /* ltimo elemento da lista */
void display_LC(struct lista_restrita **inicio_LC, /* primeiro elemento da lista */
                struct lista_restrita **final_LC); /* ltimo elemento da lista */
int lenght_LC(struct lista_restrita **inicio_LC, /* primeiro elemento da lista */
              struct lista_restrita **final_LC);  /* ltimo elemento da lista */
struct lista_restrita *encontra_objeto_LC(int posicao,
                                          struct lista_restrita **inicio_LC);
void apaga_registro_LC(struct lista_restrita *registro,
                       struct lista_restrita **inicio_LC,
                       struct lista_restrita **final_LC);
void constroi_solucao_parcialmente_gulosa(
        int n,
        int q,
        int *s,
        int *d,
        double **c,
        int penalidade,
        float *fo,
        int *inv,
        double alfa_GRASP);

int menu_solucao_inicial(void);
int menu_descida(void);
int menu_principal(void);

void descida1opt(int *s,
                 int pos,
                 double **c);
void descida2opt(int n,
                 int *s,
                 int *d,
                 double **c,
                 int q,
                 int penalidade);
void descida3opt(int n,
                 int *s,
                 int *d,
                 double **c,
                 int q,
                 int penalidade);
void vnd(int n,
         int *s,
         int pos,
         int *d,
         double **c,
         int q,
         int penalidade,
         float *fo);

void vizqq1opt(int *s, int *pos);
void vizqq2opt(int *s, int *pos);
void vizqq3opt(int *s, int *pos);
void vizqq4opt(int *s, int *pos);
void vizqq5opt(int *s, int *pos);
void vizqq6opt(int *s, int *pos);
void vns(int n,
         int *s,
         int *d,
         double **c,
         int q,
         int penalidade,
         float *fo,
         double tempo);

//---------------------------------------------------------------------------

#pragma argsused
int main(int argc, char **argv)
{
  int n;           // numero de cidades
  int dx, dy;      // localizacao do deposito (x,y)
  int q;           // capacidade dos veiculos
  int *s;          // vetor solucao corrente
  int *s_star;     // vetor melhor solucao
  int *d;          // vetor de demanda de cada cidade
  double **c;      // matriz de distancias entre as cidades
  float fo;          // funcao objetivo corrente
  float fo_star;     // melhor funcao objetivo
  int inv;         // quantidade de inviabilidade da solucao corrente
  int penalidade;  // penalidade por superar a capacidade da mochila

  int choice;

  struct lista_restrita *inicio_LC, *final_LC, *candidato;
  int tamanho_LC,     // tamanho da lista de candidatos
      tamanho_LRC;    // tamanho da lista restrita de candidatos
  double alfa_GRASP;  // percentual da LC a ser considerado

  srand(1000); // pega o numero 1000 como semente de numeros aleatorios
//  srand((unsigned) time(NULL)); // pega a hora do relogio como semente

  obter_parametros_prob("c50info.txt", &n, &q, &dx, &dy);

  s = cria_vetor(2*n+1); // vetor solucao
  s_star = cria_vetor(2*n+1); // melhor solucao encontrada
  d = cria_vetor(n+1); // vetor de demanda de cada cidade
  c = cria_matriz(n+1,n+1); // matriz de adjacencia das cidades

  le_arq_matriz("c50dados.txt", n, dx, dy, c, d);

  penalidade = calcula_penalidade(d, n);
  fo_star = - INT_MAX;

  do {
    choice = menu_principal();
    switch (choice) {
    case 1: /* Geracao de uma solucao inicial */
           switch(menu_solucao_inicial()) {
           case 1: /* Geracao gulosa de uma solucao inicial */
                 constroi_solucao_parcialmente_gulosa(n, q, s, d, c, penalidade, &fo, &inv, 0);
                 printf("Solucao construida de forma gulosa:\n");
                 imprime_solucao(s);//******************************************************************
                 printf("funcao objetivo = %f\n",fo);
	         break;
           case 2: /* Geracao parcialmente gulosa de uma solucao inicial */
                 constroi_solucao_parcialmente_gulosa(n, q, s, d, c, penalidade, &fo, &inv, 0.2);
                 printf("Solucao construida de forma gulosa:\n");
                 imprime_solucao(s);//******************************************************************
                 printf("funcao objetivo = %f\n",fo);
	         break;
           case 3: /* Geracao parcialmente gulosa de uma solucao inicial */
                 constroi_solucao_parcialmente_gulosa(n, q, s, d, c, penalidade, &fo, &inv, 1);
                 printf("Solucao construida de forma parcialmente gulosa:\n");
                 imprime_solucao(s);
                 printf("funcao objetivo = %f\n",fo);
	         break;
           }
           break;
    case 2: /* VNS */
          if (fo != 0) {
                 vns(n, s, d, c, q, penalidade, &fo, 180);
                 //printf("Solucao construida usando VNS:\n");

                 //imprime_solucao(s);
                 //printf("funcao objetivo = %f \n",fo);
          }
          else {
            printf("\n ----- Gere uma solucao inicial -------\n");
            getchar();
          }
          break;
    case 3: /* VND */
          if (fo != 0) {
                 vnd(n, s, 2, d, c, q, penalidade, &fo);
                 printf("Solucao construida usando VND:\n");

                 imprime_solucao(s);
                 printf("funcao objetivo = %f \n",fo);
          }
          else {
            printf("\n ----- Gere uma solucao inicial -------\n");
            getchar();
          }
          break;

   default:
          break;
    }
  } while (choice != 0);


  libera_vetor(s);

  libera_vetor(s_star);
  libera_vetor(d);
  libera_matriz(c,n);
  return 0;
} /* fim do programa principal */

//---------------------------------------------------------------------------

/* cria memoria para um vetor de tam posicoes */
int *cria_vetor(int tam)
{
  int *vetor;

  vetor = (int *) malloc(tam*sizeof(int));
  if (!vetor){
    printf("Falta memoria para alocar o vetor de ponteiros");
    exit(1);
  }
  return vetor;
}
/* cria memoria para um vetor de tam posicoes */
double *cria_vetor_double(int tam)
{
  double *vetor;

  vetor = (double *) malloc(tam*sizeof(double));
  if (!vetor){
    printf("Falta memoria para alocar o vetor de ponteiros");
    exit(1);
  }
  return vetor;
}
/* libera memoria de um vetor */
void libera_vetor(int *vetor)
{
  free(vetor);
}

/* Cria matriz de ponteiros para inteiros com nlinhas e ncolunas */
double **cria_matriz(int nlinhas, int ncolunas)
{
  register int i;
  double **matriz;

  matriz = (double **) malloc(nlinhas*sizeof(double *));
  if (!matriz) {
        printf("Falta memoria para alocar a matriz de ponteiros\n");
        exit(1);
  }
  for (i=0; i < nlinhas; i++) {
    matriz[i] = (double *) malloc(ncolunas*sizeof(double));
    if (!matriz[i]){
      printf("Falta memoria para alocar a matriz de ponteiros.\n");
      exit(1);
    }
  }
  return matriz;
}


void libera_matriz(double **matriz, int nlinhas)

{
  register int i;

  for (i=nlinhas-1; i >= 0; i--)
    free((int *) matriz[i]);
  free((int *) matriz);
}


/* le um arquivo no formato "Format of vehicle routing problem instance" */
/* http://ina.eivd.ch/collaborateurs/etd/problemes.dir/vrp.dir/vrp.html */
void obter_parametros_prob(char nomearq[], int *n, int *q, int *dx, int *dy)
{
  float best_sol=0;

  FILE *arquivo;

  arquivo = fopen(nomearq,"r");
  if (!arquivo) {
     printf("O Arquivo %s nao pode ser aberto.\n", nomearq);
     getchar();
     exit(1);
  }
  fscanf(arquivo, "%d %f", n, &best_sol);
  fscanf(arquivo, "%d", q);
  fscanf(arquivo, "%d %d", dx, dy);

  fclose(arquivo);
}

/* le um arquivo no formato num_cid di1 di2 ... dij demanda, onde i,j=[1..n] */
void le_arq_matriz(char nomearq[], int n, int dx, int dy, double **distancia, int *demanda)
{
  int *vet_x, *vet_y;
  int x, y, i, j,  valor;
  FILE *arquivo;

  vet_x = cria_vetor(n);
  vet_y = cria_vetor(n);
  arquivo = fopen(nomearq,"r");
  if (!arquivo) {
     printf("O Arquivo %s nao pode ser aberto.\n", nomearq);
     getchar();
     exit(1);
  }
  vet_x[0] = dx;
  vet_y [0] = dy;
  demanda[0] = 0;
  while (!feof(arquivo)){
    fscanf(arquivo, "%d %d %d %d", &i, &x, &y, &valor);
    vet_x[i] = x;
    vet_y[i] = y;
    demanda[i] = valor;
  }

  if (i != n) { // verifica a "integridade" dos dados lidos
     printf("O Arquivo %s nao corresponde as especificacoes.\n", nomearq);
     printf("O Arquivo %s possui %d regitros enquanto o problema tem %d cidades.\n", nomearq, i, n);
     getchar();  exit(1);
  }
  // gera a matriz de distancias calculado a partir das distancias euclidianas dos pontos
  for(i=0; i <= n; i++)
   {
     distancia[i][i] = 0;
     for(j=i+1; j <= n; j++)
      {
        distancia[i][j] = (double) sqrt(pow(vet_x[i]-vet_x[j],2)+pow(vet_y[i]-vet_y[j],2));
        distancia[j][i] = distancia[i][j];
      }
   }

//  fclose(arquivo);
  libera_vetor(vet_x);
  libera_vetor(vet_y);
}

/* imprime a solucao */
void imprime_solucao(int *s)
{

    printf("\n[");
    for(int j=1; j<s[0]; j++)
      {
        printf("%d-",s[j]);
      }
    printf("%d]\n",s[s[0]]);
//    getch();
}

/* calcula a funcao objetivo */
float calcula_fo(int *s, double **distancia, int inviabilidade, int penalidade)
{
    float dist_percorrida = 0;
//    printf("0");
    for ( int j=1; j < s[0]; j++ ) {
//      printf("+[%d,%d]%f",s[j],s[j+1],distancia[s[j]][s[j+1]]);
      dist_percorrida += distancia[s[j]][s[j+1]];
    }
//   printf("=%f",dist_percorrida);
//    printf("+%d*%d",penalidade,inviabilidade);
    return (dist_percorrida + penalidade * inviabilidade);
}

/* Gera numero aleatorio entre min e max */
double randomico(double min, double max)
{
  if (min == max) return min;
  return ((double) (rand()%10000/10000.0)*(max-min) + min);
}

/* atualiza a melhor solucao */
void atualiza_melhor_solucao(int *s, int *s_star, int num_objetos)
{
   for (int j=0; j<num_objetos; j++) s_star[j] = s[j];
}

/* calcula a penalidade por exceder a capacidade do veiculo */
int calcula_penalidade(int *d, int n)
{
    int penalidade = 0;
    for (int j=1; j<=n; j++) {
        penalidade += d[j];
      }
    return penalidade;
}

/* calcula a inviabilidade por exceder a capacidade do veiculo */
int calcula_inviabilidade(int *s, int *d, int q)
{
    int inviabilidade = 0;
    int cont=0;
    for (int j=1; j<=s[0]; j++) {
     if (s[j] == 0) {
        if (cont > q)
           inviabilidade += cont - q;

      cont=0;
      }
     else
        cont += d[s[j]];
    }
    return inviabilidade;
}

void inicializa_vetor(int *vetor, int tam)
{
    for (int j=0; j<tam; j++) vetor[j] = 0;
}

/* Menu de geracao de uma solucao inicial */
int menu_solucao_inicial(void)
{
  int escolha;

  do {
    printf("\nGERACAO DA SOLUCAO INICIAL: \n");
    printf("                  1. Gulosa \n");
    printf("                  2. Parcialmente gulosa (GRASP) \n");
    printf("                  3. Aleatoria \n");
    printf("                  escolha: ");
    scanf("%d",&escolha);
  } while (escolha < 1 || escolha > 3);
  return escolha;
}

/* Insere um registro de forma ordenada em uma lista duplamente encadeada ordenada */
void insere_LC(struct lista_restrita *i,    /* novo elemento */
               struct lista_restrita **inicio_LC, /* primeiro elemento da lista */
               struct lista_restrita **final_LC)  /* ltimo elemento da lista */
{
  struct lista_restrita *antigo, *ponteiro;

  if (*final_LC==NULL) {  /* primeiro elemento da lista */
     i->proximo = NULL;
     i->anterior = NULL;
     *final_LC = i;
     *inicio_LC = i;
     return;
  }
  ponteiro = *inicio_LC; /* comea no topo da lista */
  antigo = NULL;
  while(ponteiro) {
    if(ponteiro->distancia < i->distancia) {
      antigo = ponteiro;
      ponteiro = ponteiro->proximo;
    }
    else {
      if(ponteiro->anterior) {
        ponteiro->anterior->proximo = i;
        i->proximo = ponteiro;
        i->anterior = ponteiro->anterior;
        ponteiro->anterior = i;
        return;
      }
      i->proximo = ponteiro; /* novo primeiro elemento */
      i->anterior = NULL;
      ponteiro->anterior = i;
      *inicio_LC = i;
      return;
    }
  }
  antigo->proximo = i; /* coloca no final */
  i->proximo = NULL;
  i->anterior = antigo;
  *final_LC = i;
}


/* Mostra o conteudo da lista completa */
void display_LC(struct lista_restrita **inicio_LC, /* primeiro elemento da lista */
	        struct lista_restrita **final_LC)  /* ltimo elemento da lista */
{
  struct lista_restrita *registro = *inicio_LC;
  int cont = 0;

  while (registro) {
    printf("LC[%2d]   cidade = %2d  distancia = %2f demanda = %2d \n",
            cont, registro->cidade, registro->distancia, registro->demanda);
    registro = registro->proximo;   /* obtem prximo endereo */
    cont++;
  }
  printf("\n");
}


/* Mostra o comprimento da lista completa */
int lenght_LC(struct lista_restrita **inicio_LC, /* primeiro elemento da lista */
              struct lista_restrita **final_LC)  /* ltimo elemento da lista */
{
  struct lista_restrita *registro = *inicio_LC;
  int cont = 0;

  while(registro) {
    registro = registro->proximo;   /* obtem prximo endereo */
    cont++;
  }
  return cont;
}


/* Encontra o registro de numero posicao */
struct lista_restrita *encontra_objeto_LC(int posicao,
                                          struct lista_restrita **inicio_LC)
{
  struct lista_restrita *registro = *inicio_LC;
  int contador = 0;

  while (registro) {
    if (contador == posicao) return registro;
    contador++;
    registro = registro->proximo;  /* obtem novo registro */
  }
  printf("Registro nao encontrado.\n");
  return NULL; /* nao encontrou */
}


/* Remove o elemento registro da lista */
void apaga_registro_LC(struct lista_restrita *registro,
                       struct lista_restrita **inicio_LC,
                       struct lista_restrita **final_LC)
{
  if (registro) {
    if (*inicio_LC==registro) {
       *inicio_LC=registro->proximo;
       if (*inicio_LC)
          (*inicio_LC)->anterior = NULL;
       else *final_LC = NULL;
    }
    else {
      registro->anterior->proximo = registro->proximo;
      if (registro!=*final_LC)
         registro->proximo->anterior = registro->anterior;
      else
         *final_LC = registro->anterior;
    }
    free(registro); /* devolve memria para o sistema */
  }
}

void constroi_solucao_parcialmente_gulosa(
        int n,
        int q,
        int *s,
        int *d,
        double **c,
        int penalidade,
        float *fo,
        int *inv,
        double alfa_GRASP)
{
    struct lista_restrita *inicio_LC, *final_LC, *candidato;
    struct lista_restrita *inicio_LC_aux, *final_LC_aux, *candidato_aux;
    int posicao_escolhida; // objeto escolhido
    int fo_star;
    int tamanho_LC,
        tamanho_LRC;
    int q_travel;
    int demanda;
    int cidade;

    /* Inicio da Fase de Construcao de uma solucao */
    printf("\n Construindo nova solucao ...\n");
    inicio_LC = final_LC = NULL;
    inicializa_vetor(s,2*n+1);
    *inv = 0;
    *fo = 0;
    q_travel = 0;
    s[0]++; // inicia no deposito
//  s[1] = 0;
    tamanho_LC = 0;
    for (int j=1; j<=n; j++){
      candidato = (struct lista_restrita *)malloc(sizeof(struct lista_restrita));
      if (!candidato) {
         printf("Faltando memoria ...\n");
         exit(1);
      }
      /* vou inserir um registro em uma lista de candidatos,
         ordenada pela distancia */
      candidato->cidade = j;
      candidato->demanda = d[j];
      candidato->distancia = c[0][j];
      insere_LC(candidato, &inicio_LC, &final_LC);
      tamanho_LC++;
//    printf("Tamanho da LC = %2d\n", tamanho_LC);
//    display_LC(&inicio_LC, &final_LC);
//    getchar();
    }
    printf("Tamanho da LC = %2d\n", tamanho_LC);
    display_LC(&inicio_LC, &final_LC);
//    getchar();

    while (tamanho_LC > 0){
       tamanho_LRC = MAX(1,(int) (alfa_GRASP * tamanho_LC));
       printf("Tamanho da LRC = %2d \n",tamanho_LRC);
       posicao_escolhida = random(tamanho_LRC);
       printf("Posicao escolhida da LRC = %2d \n",posicao_escolhida);
       candidato = encontra_objeto_LC(posicao_escolhida, &inicio_LC);
       printf("A cidade escolhida = %2d \n",candidato->cidade);
       if ((q_travel) + d[candidato->cidade] <= q){
         s[0]++;
         s[s[0]] = candidato->cidade;
         q_travel += d[candidato->cidade];
         printf("Cidade %2d de demanda %2d incluida na solucao \n",
                 candidato->cidade, d[candidato->cidade]);
         printf("Demanda atual atendida neste tour = %2d \n",q_travel);
         apaga_registro_LC(candidato, &inicio_LC, &final_LC);
         tamanho_LC--;
//         display_LC(&inicio_LC, &final_LC);
       }
       else {
         printf("A inclusao da cidade %2d de demanda %2d supera capacidade do veiculo \n",
                 candidato->cidade, d[candidato->cidade]);
         printf("Vou retornar o veiculo ao deposito e incluir o cidade %2d em um  novo tour\n",candidato->cidade);
//       getchar();
       s[0]++;
       s[s[0]] = 0; // insere volta ao deposito
       q_travel = 0;
       s[0]++;
       s[s[0]] = candidato->cidade;
       q_travel += d[candidato->cidade];
       apaga_registro_LC(candidato, &inicio_LC, &final_LC);
       tamanho_LC--;
//       display_LC(&inicio_LC, &final_LC);
        }
    //Atualiza os candidatos
    inicio_LC_aux = final_LC_aux = NULL;
    for (int j=1; j<=tamanho_LC; j++){
      candidato = encontra_objeto_LC(0, &inicio_LC);
      demanda=candidato->demanda;
      cidade=candidato->cidade;
      apaga_registro_LC(candidato, &inicio_LC, &final_LC);

      /* vou inserir um registro atualizado em uma nova lista de
         candidatos, ordenada pela distancia */
      candidato = (struct lista_restrita *)malloc(sizeof(struct lista_restrita));
      candidato->demanda=demanda;
      candidato->cidade=cidade;
//      printf("demanda=%d cidade=%d", candidato->demanda, candidato->cidade);
      candidato->distancia = c[s[s[0]]][candidato->cidade];
      insere_LC(candidato, &inicio_LC_aux, &final_LC_aux);
    }
      inicio_LC = inicio_LC_aux;
      final_LC = final_LC_aux;
//      printf("Tamanho da LC = %2d\n", tamanho_LC);
//      display_LC(&inicio_LC, &final_LC);
//      getchar();
    } // while
    s[0]++; // retorna ao deposito
//    s[s[0]] = 0;
    /* Fim da Fase de Construcao de uma solucao */

    *fo = calcula_fo(s, c, *inv, penalidade);
}

int menu_descida(void)
{
  int escolha;

  do {
    printf("\nMETODO DA DESCIDA: \n");
    printf("                  1. Metodo da Descida 1 Optimal \n");
    printf("                  2. Metodo da Descida 2 Optimal \n");
    printf("                  escolha: ");
    scanf("%d",&escolha);
  } while (escolha < 1 || escolha > 2);
  return escolha;
}

/* Menu principal */
int menu_principal(void)
{
  int escolha;

  do {
    printf("\n************************************************* \n");
    printf("         1. Gere solucao inicial \n");
    printf("         2. VNS \n");
    printf("         3. VND \n");
    printf("         0. Sair \n");
    printf("         Escolha: ");
    scanf("%d",&escolha);
  } while (escolha < 0 || escolha > 10);
  return escolha;
}

void descida1opt(int *s, int pos, double **c)
{ int a, b, aux;
  for(a=pos ; s[a]!=0 ; a--);
  for(b=pos ; s[b]!=0 ; b++);

  double dist = 0;

  for(int i=a ; i<b; i++)
   dist +=c[s[i]][s[i+1]];
  int mi=a;
  int mj=a;
  double distl;
  for(int i=a+1 ; i<b ; i++)
    for(int j=i ; j<b ; j++) {
      aux = s[j];
      s[j] = s[i];
      s[i] = aux;
      distl = 0;
      for(int k=a ; k<b; k++)
        distl +=c[s[k]][s[k+1]];
      if(distl<dist){
        mi = i;
        mj = j;
        dist = distl;
      }
      aux = s[j];
      s[j] = s[i];
      s[i] = aux;
    }
    aux = s[mj];
    s[mj] = s[mi];
    s[mi] = aux;

/*    printf("-------------------------------------------------\n");
    printf("Melhor solucao encontrada pelo Metodo da descida 1 Optimal:\n");
    printf("Funcao objetivo: %f\n", *fo);
    printf("Peso da mochila: %d    Inviabilidade = %d \n", *peso_objetos, MAX(0,*peso_objetos - b));
    printf("-------------------------------------------------\n");*/
}//descida1opt

void descida2opt(int n, int *s, int *d, double **c, int q, int penalidade)
{
  int a=1;
  int b=s[0];
  int aux, inv, fo, fosl;

  inv = calcula_inviabilidade(s,d,q);
  fo = calcula_fo(s, c, inv, penalidade);
  int mi=a; int mj=a;
  for(int i=a+1 ; i<b ; i++)
    for(int j=i ; j<b ; j++) {
      aux = s[j];
      s[j] = s[i];
      s[i] = aux;

      inv = calcula_inviabilidade(s,d,q);
      fosl = calcula_fo(s, c, inv, penalidade);

      if(fosl<fo){
        mi = i;
        mj = j;
        fo = fosl;
      }
      aux = s[j];
      s[j] = s[i];
      s[i] = aux;
    }
    aux = s[mj];
    s[mj] = s[mi];
    s[mi] = aux;

/*    printf("-------------------------------------------------\n");
    printf("Melhor solucao encontrada pelo Metodo da descida 2 Optimal:\n");
    printf("Funcao objetivo: %f\n", *fo);
    printf("Peso da mochila: %d    Inviabilidade = %d \n", *peso_objetos, MAX(0,*peso_objetos - b));
    printf("-------------------------------------------------\n"); */
}//descida2opt

void descida3opt(int n, int *s, int *d, double **c, int q, int penalidade)
{
  int a=1;
  int b=s[0];
  int aux, inv, fo, fosl;

  inv = calcula_inviabilidade(s,d,q);
  fo = calcula_fo(s, c, inv, penalidade);
  int mi=a; int mj=a; int mk=a;
  for(int i=a+1 ; i<b ; i++)
    for(int j=i ; j<b ; j++)
      for(int k=j ; k<b ; k++) {
      aux = s[i];
      s[i] = s[j];
      s[j] = s[k];
      s[k] = aux;

      inv = calcula_inviabilidade(s,d,q);
      fosl = calcula_fo(s, c, inv, penalidade);

      if(fosl<fo){
        mi = i;
        mj = j;
        mk = k;
        fo = fosl;
      }
      aux = s[i];
      s[i] = s[j];
      s[j] = s[k];
      s[k] = aux;    }
    aux = s[mi];
    s[mi] = s[mj];
    s[mj] = s[mk];
    s[mk] = aux;

/*    printf("-------------------------------------------------\n");
    printf("Melhor solucao encontrada pelo Metodo da descida 3 Optimal:\n");
    printf("Funcao objetivo: %f\n", *fo);
    printf("Peso da mochila: %d    Inviabilidade = %d \n", *peso_objetos, MAX(0,*peso_objetos - b));
    printf("-------------------------------------------------\n"); */
}//descida3opt

void vnd(int n, int *s, int pos, int *d, double **c, int q, int penalidade, float *fo)
{
    int *sl;
    float fosl;
    int inv = 0;
    sl = cria_vetor(2*n+1);

    for (int i = 0; i <= 2*n; i++) sl[i] = s[i];
    fosl = *fo;

    int k = 1;

    while(k <= 2) //while 1
    {
       switch(k){
       case 1: descida1opt(sl, pos, c);
               break;
       case 2: descida2opt(n, sl, d, c, q, penalidade);
               break;
       case 3: descida3opt(n, sl, d, c, q, penalidade);
               break;
       }
       inv = calcula_inviabilidade(sl,d,q);
       fosl = calcula_fo(sl, c, inv, penalidade);
       if (fosl < *fo)
       {
          for (int i = 0; i <= 2*n; i++) s[i] = sl[i];
          *fo = fosl;
          k = 1;
       }
       else k++;
    }//while 1

/*
    inv = calcula_inviabilidade(s,d,q);
    printf("-------------------------------------------------\n");
    printf("Melhor solucao encontrada pelo VND:\n");
    printf("Funcao objetivo: %f\n", *fo);
    printf("Inviabilidade total de: %d\n", inv);
    printf("-------------------------------------------------\n");
    imprime_solucao(s);
*/

    libera_vetor(sl);
}//vnd


void vizqq1opt(int *s, int *pos)
{
  *pos = random(s[0]-3)+2;

  while ( s[*pos] == 0 ) {
    *pos = random(s[0]-3)+2;
  }
  int i, j;
  for(i=*pos ; s[i]!=0 ; i--);
  for(j=*pos ; s[j]!=0 ; j++);

  int vizqq = random(j-i-2)+i+1;

  if ( (j - i == 1) || (j - i == 0)  )
    while ( vizqq == *pos ) {
      vizqq = random(j-i-2)+i+1;
    }

  int aux;
  aux = s[*pos];
  s[*pos] = s[vizqq];
  s[vizqq] = aux;
}//vizqq1opt

void vizqq2opt(int *s, int *pos)
{
  *pos = random(s[0]-3)+2;
  int vizqq = random(s[0]-3)+2;

  int aux;
  aux = s[*pos];
  s[*pos] = s[vizqq];
  s[vizqq] = aux;
}//vizqq2opt

void vizqq3opt(int *s, int *pos)
{
  *pos = random(s[0]-3)+2;
  int vizqq1 = random(s[0]-3)+2;
  int vizqq2 = random(s[0]-3)+2;

  int aux;
  aux = s[*pos];
  s[*pos] = s[vizqq1];
  s[vizqq1] = s[vizqq2];
  s[vizqq2] = aux;
}//vizqq3opt

void vizqq4opt(int *s, int *pos)
{
  *pos = random(s[0]-3)+2;
  int vizqq1 = random(s[0]-3)+2;
  int vizqq2 = random(s[0]-3)+2;
  int vizqq3 = random(s[0]-3)+2;

  int aux;
  aux = s[*pos];
  s[*pos] = s[vizqq1];
  s[vizqq1] = s[vizqq2];
  s[vizqq2] = s[vizqq3];
  s[vizqq3] = aux;
}//vizqq4opt

void vizqq5opt(int *s, int *pos)
{
  *pos = random(s[0]-3)+2;
  int vizqq1 = random(s[0]-3)+2;
  int vizqq2 = random(s[0]-3)+2;
  int vizqq3 = random(s[0]-3)+2;
  int vizqq4 = random(s[0]-3)+2;

  int aux;
  aux = s[*pos];
  s[*pos] = s[vizqq1];
  s[vizqq1] = s[vizqq2];
  s[vizqq2] = s[vizqq3];
  s[vizqq3] = s[vizqq4];
  s[vizqq4] = aux;
}//vizqq5opt

void vizqq6opt(int *s, int *pos)
{
  *pos = random(s[0]-3)+2;
  int vizqq1 = random(s[0]-3)+2;
  int vizqq2 = random(s[0]-3)+2;
  int vizqq3 = random(s[0]-3)+2;
  int vizqq4 = random(s[0]-3)+2;
  int vizqq5 = random(s[0]-3)+2;

  int aux;
  aux = s[*pos];
  s[*pos] = s[vizqq1];
  s[vizqq1] = s[vizqq2];
  s[vizqq2] = s[vizqq3];
  s[vizqq3] = s[vizqq4];
  s[vizqq4] = s[vizqq5];
  s[vizqq5] = aux;
}//vizqq6opt


void vns(int n, int *s, int *d, double **c, int q, int penalidade, float *fo, double tempo)
{
    int pos;
    int *sl;
    float fosl;
    int inv = 0;
    sl = cria_vetor(2*n+1);
    clock_t inicioCPU, fimCPU;
    inicioCPU = fimCPU = clock();

    while((fimCPU - inicioCPU) / CLK_TCK < tempo) //while1
    {
       int k = 1;
       while(k <= 12) //while2
       {
          for (int i = 0; i <= 2*n; i++) sl[i] = s[i];
          fosl = *fo;
          switch(k){ //Gere um vizinho qualquer na vizinhanca corrente
          case 1: vizqq1opt(sl, &pos);
                  break;
          case 2: vizqq2opt(sl, &pos);
                  break;
          case 3: vizqq3opt(sl, &pos);
                  break;
          case 4: vizqq4opt(sl, &pos);
                  break;
          case 5: vizqq4opt(sl, &pos);
                  break;
          case 6: vizqq5opt(sl, &pos);
                  break;
          case 7: vizqq5opt(sl, &pos);
                  break;
          case 8: vizqq5opt(sl, &pos);
                  break;
          case 9: vizqq6opt(sl, &pos);
                  break;
          case 10: vizqq6opt(sl, &pos);
                  break;
          case 11: vizqq6opt(sl, &pos);
                  break;
          case 12: vizqq6opt(sl, &pos);
                  break;
          }
          inv = calcula_inviabilidade(sl,d,q);
          fosl = calcula_fo(sl, c, inv, penalidade);
          vnd(n, sl, pos, d, c, q, penalidade, &fosl);
          if (fosl < *fo)
          {
             for (int i = 0; i <= 2*n; i++) s[i] = sl[i];
             *fo = fosl;
             k = 1;

             inv = calcula_inviabilidade(s,d,q);
             printf("\n-------------------------------------------------\n");
             printf("Melhor solucao encontrada ate o momento pelo VNS:\n");
             printf("Funcao objetivo: %f\n", *fo);
             printf("Inviabilidade total de: %d\n", inv);
             imprime_solucao(s);
             printf("-------------------------------------------------\n");

          }
          else k++;
       }//while2
       fimCPU = clock();
    }//while 1
    inv = calcula_inviabilidade(s,d,q);
    printf("-------------------------------------------------\n");
    printf("Melhor solucao encontrada pelo VNS em %0.f segundos:\n", tempo);
    printf("Funcao objetivo: %f\n", *fo);
    printf("Inviabilidade = %d \n", inv);
    imprime_solucao(s);
    printf("-------------------------------------------------\n");
    libera_vetor(sl);
}//vns

